# (c) 2024 Niels Provos
#

import io
from functools import wraps
import cv2
import time
import base64
import torch
import numpy as np
from PIL import Image, ImageDraw
from pathlib import Path
import hashlib

from . import constants as C


def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time_ms = (end_time - start_time) * 1000.0
        print(f"Function {func.__name__} took {total_time_ms:.1f} ms")
        return result

    return timeit_wrapper


def filename_add_version(filename):
    filename = Path(filename)
    last_component = filename.stem.split("_")[-1]
    if last_component.startswith("v"):
        stem = "_".join(filename.stem.split("_")[:-1])
        version = int(last_component[1:])
        version += 1
        image_filename = f"{stem}_v{version}.png"
    else:
        image_filename = f"{filename.stem}_v2.png"

    return str(filename.parent / image_filename)


def filename_previous_version(filename):
    filename = Path(filename)
    last_component = filename.stem.split("_")[-1]
    if not last_component.startswith("v"):
        return None

    stem = "_".join(filename.stem.split("_")[:-1])
    version = int(last_component[1:])
    version -= 1
    if version > 1:
        image_filename = f"{stem}_v{version}.png"
    else:
        image_filename = f"{stem}.png"

    return str(filename.parent / image_filename)


@timeit
def to_image_url(img_data):
    """Converts an image to a data URL."""
    if not isinstance(img_data, Image.Image):
        img_data = Image.fromarray(img_data)
    buffered = io.BytesIO()
    img_data.save(buffered, format="PNG")
    return to_data_url(buffered.getvalue())


def to_data_url(data):
    """Converts binary data to a data URL."""
    url_str = base64.b64encode(data).decode("utf-8")
    return f"data:image/png;base64,{url_str}"


def torch_get_device():
    """
    Returns the appropriate torch device based on the availability of CUDA or MPS.

    Returns:
        torch.device: The torch device (cuda, mps, or cpu) based on availability.
    """
    if torch.cuda.is_available():
        return torch.device("cuda")
    elif torch.backends.mps.is_available():
        return torch.device("mps")
    return torch.device("cpu")


def find_bounding_box(mask_image, padding=50):
    """
    Finds the bounding box of a given mask image.

    Args:
        mask_image (PIL.Image.Image or np.array): The mask image.
        padding (int, optional): The padding to apply to the bounding box. Defaults to 50.

    Returns:
        tuple: A tuple containing the coordinates of the bounding box in the format (xmin, ymin, xmax, ymax).
    """
    if isinstance(mask_image, Image.Image):
        mask_array = np.array(mask_image)
    else:
        mask_array = mask_image

    nonzero_y, nonzero_x = np.nonzero(mask_array > 0)
    if len(nonzero_x) == 0 or len(nonzero_y) == 0:
        return (0, 0, 0, 0)
    xmin, xmax = nonzero_x.min(), nonzero_x.max()
    ymin, ymax = nonzero_y.min(), nonzero_y.max()
    # Apply padding to the bounding box
    xmin -= padding
    xmax += padding
    ymin -= padding
    ymax += padding
    return (xmin, ymin, xmax, ymax)


def find_square_from_bounding_box(xmin, ymin, xmax, ymax):
    """
    Finds a square from a given bounding box.

    Args:
        xmin (int): The minimum x-coordinate of the bounding box.
        ymin (int): The minimum y-coordinate of the bounding box.
        xmax (int): The maximum x-coordinate of the bounding box.
        ymax (int): The maximum y-coordinate of the bounding box.

    Returns:
        tuple: A tuple containing the coordinates of the square in the format (x1, y1, x2, y2).
    """
    width = xmax - xmin
    height = ymax - ymin
    size = max(width, height)
    xcenter = (xmin + xmax + 1) // 2
    ycenter = (ymin + ymax + 1) // 2
    x1 = xcenter - size // 2
    y1 = ycenter - size // 2
    x2 = xcenter + size // 2
    y2 = ycenter + size // 2
    return (x1, y1, x2, y2)


def move_bounding_box_to_image(image, bounding_box):
    width, height = image.size
    xmin, ymin, xmax, ymax = bounding_box
    if xmin < 0:
        xmax -= xmin
        xmin = 0
    if ymin < 0:
        ymax -= ymin
        ymin = 0
    if xmax >= width:
        xmin -= xmax - width
        xmax = width
    if ymax >= height:
        ymin -= ymax - height
        ymax = height

    # make sure the bounding box is within the image
    # this will change the aspect ratio of the bounding box
    # but we will resize the image to square anyway
    if xmin < 0:
        xmin = 0
    if ymin < 0:
        ymin = 0

    return (xmin, ymin, xmax, ymax)


def find_square_bounding_box(mask_image, padding=50):
    """
    Finds the square bounding box for a given mask image.

    Args:
        mask_image: The mask image for which the square bounding box needs to be found.
        padding: The padding to apply to the bounding box. Defaults to 50.

    Returns:
        The square bounding box that fits the mask image.
    """
    bounding_box = find_bounding_box(mask_image, padding=padding)
    square_box = find_square_from_bounding_box(*bounding_box)
    fit_box = move_bounding_box_to_image(mask_image, square_box)
    return fit_box


def image_overlay(image, segmented_image):
    """
    Blend the main image with a segmented image using PIL.

    Args:
        image (PIL.Image.Image): The main image in RGB format.
        segmented_image (str or PIL.Image.Image): The segmentation map in RGB format.

    Returns:
        PIL.Image.Image: A new PIL Image object with the result of the blending.
    """
    alpha_segmented = 0.8  # transparency for the segmentation map
    result = Image.blend(image, segmented_image, alpha_segmented)

    return result


def apply_color_tint(img, color, alpha):
    # Create an overlay image filled with the specified color
    overlay = Image.new("RGB", img.size, color=color)

    # Blend the original image with the overlay
    return Image.blend(img.convert("RGB"), overlay, alpha)


def find_pixel_from_click(img_data, x, y, width, height):
    """Find the pixel coordinates in the image from the click coordinates."""
    img_width, img_height = img_data.size
    x_ratio = img_width / width
    y_ratio = img_height / height
    return int(x * x_ratio), int(y * y_ratio)


def find_pixel_from_event(state, e, rect_data):
    clientX = e["clientX"]
    clientY = e["clientY"]

    rectTop = rect_data["top"]
    rectLeft = rect_data["left"]
    rectWidth = rect_data["width"]
    rectHeight = rect_data["height"]

    x = clientX - rectLeft
    y = clientY - rectTop

    return find_pixel_from_click(state.imgData, x, y, rectWidth, rectHeight)


def feather_mask(mask, num_expand=50):
    """
    Expand and feather a mask.

    Args:
        mask (numpy.ndarray): The input mask.
        num_expand (int, optional): The number of times to expand the mask. Defaults to 50.

    Returns:
        numpy.ndarray: The expanded and feathered mask.
    """
    was_pil_image = False
    if isinstance(mask, Image.Image):
        mask = np.array(mask)
        was_pil_image = True

    # Expand the mask
    kernel = np.ones((num_expand, num_expand), np.uint8)
    expanded_mask = cv2.dilate(mask, kernel, iterations=1)

    # Feather the expanded mask
    feathered_mask = cv2.GaussianBlur(
        expanded_mask, (num_expand * 2 + 1, num_expand * 2 + 1), 0
    )

    if was_pil_image:
        feathered_mask = Image.fromarray(feathered_mask)

    return feathered_mask


def postprocess_depth_map(depth_map, image_alpha, final_blur=5):
    """
    Apply edge extension technique to extend the depthmap beyond the edge of the alpha channel.

    Args:
        depth_map (numpy.ndarray): The depth map.
        image_alpha (numpy.ndarray): The alpha channel of the image.
        final_blur (int): The amount of post blur to apply to the depth map

    Returns:
        numpy.ndarray: The post-processed depth map.

    """
    depth_map[image_alpha != 255] = 0

    kernel = np.ones((3, 3), np.uint8)

    # erode the alpha channel to remove the feathering
    image_alpha = cv2.erode(image_alpha, kernel, iterations=3)

    depth_map_blur = cv2.blur(depth_map, (15, 15))
    depth_map[image_alpha != 255] = depth_map_blur[image_alpha != 255]

    for i in range(20):
        depth_map_dilated = cv2.dilate(depth_map, kernel, iterations=1)
        depth_map[image_alpha != 255] = depth_map_dilated[image_alpha != 255]

    # make final blur an odd number - required by GaussianBlur
    if final_blur % 2 == 0:
        final_blur += 1
    depth_map = cv2.GaussianBlur(depth_map, (final_blur, final_blur), 0)

    # normalize to the smallest value
    smallest_vale = int(np.quantile(depth_map[image_alpha == 255], 0.01))
    smallest_vale = int(smallest_vale)

    # change dtype of depth map to int
    depth_map = depth_map.astype(np.int16)
    depth_map[:, :] -= smallest_vale
    depth_map = np.clip(depth_map, 0, 255)
    depth_map = depth_map.astype(np.uint8)

    return depth_map


def premultiply_alpha_numpy(img):
    """
    Premultiplies the alpha channel of an RGBA image using NumPy.

    Args:
        img (PIL.Image.Image): The input RGBA image.

    Returns:
        PIL.Image.Image: The premultiplied RGBA image.
    """
    # Convert the image to a NumPy array
    arr = np.array(img)

    # Split the array into RGB and Alpha components
    rgb = arr[..., :3]
    a = arr[..., 3:] / 255.0  # Normalise alpha to (0, 1)

    # Perform premultiplication
    premultiplied_rgb = (rgb * a).astype(np.uint8)

    # Combine the new RGB with the original alpha
    premultiplied = np.dstack((premultiplied_rgb, arr[..., 3]))

    # Convert back to an Image
    return Image.fromarray(premultiplied, "RGBA")


def draw_circle(
    image, center, radius, fill_color=(255, 128, 128), outline_color=(255, 55, 55)
):
    """
    Draw a circle on a PIL Image.

    Args:
        image (PIL.Image.Image): The PIL Image object where the circle will be drawn.
        center (tuple): The center coordinates of the circle in the format (x, y).
        radius (int): The radius of the circle.
        fill_color (tuple): The color of the circle's interior in RGB format.
        outline_color (tuple): The color of the circle's outline in RGB format.
    """
    draw = ImageDraw.Draw(image)
    left = center[0] - radius
    top = center[1] - radius
    right = center[0] + radius
    bottom = center[1] + radius
    draw.ellipse([left, top, right, bottom], fill=fill_color, outline=outline_color)


def get_gltf_iframe(gltf_uri):
    return f"""
    <html>
    <head>
        <script type="module" src="https://ajax.googleapis.com/ajax/libs/model-viewer/3.4.0/model-viewer.min.js"></script>
        <style>
            body, html {{
                margin: 0;
                padding: 0;
                width: 100%;
                height: 100%;
            }}
            model-viewer {{
                width: 100%;
                height: 100%;
            }}
        </style>
    </head>
    <body>
    <model-viewer
        id="{C.IFRAME_MODEL_VIEWER}"
        src="{gltf_uri}"
        alt="glTF Scene"
        ar
        autoRotate
        camera-target="0m 0m 0m"
        camera-orbit="3.106650330236851rad 1.5658376358588284rad 50m"
        field-of-view="8"
        min-camera-orbit='auto auto 1%'
        max-camera-orbit='auto auto 100%'
        min-field-of-view='1deg'
        max-field-of-view='60deg'
        camera-controls touch-action="pan-y"
        style="width: 100%; height: 100vh;"
    ></model-viewer>
    </body>
    </html>
    """


def get_no_gltf_available():
    return """
    <html>
    <head>
        <style>
            body, html {
                margin: 0;
                padding: 0;
                width: 100%;
                height: 100%;
                display: flex;
                justify-content: center;
                align-items: center;
            }
            h1 {
                font-family: sans-serif;
                font-size: 24px;
                color: #333;
            }
        </style>
    </head>
    <body>
        <h1>No glTF file available.</h1>
    </body>
    </html>
    """


def highlight_selected_element(
    classnames, index, highlight_class="color-is-selected-light"
):
    """
    Highlights the selected element in a list of classnames by adding a highlight class.

    Args:
        classnames (list): List of classnames.
        index (int): Index of the element to be highlighted.
        highlight_class (str, optional): The highlight class to be added. Defaults to 'color-is-selected-light'.

    Returns:
        list: A new list of classnames with the selected element highlighted.
    """
    new_classnames = []
    selected_background = f" {highlight_class}"
    for i, classname in enumerate(classnames):
        classname = classname.replace(selected_background, "")
        if i == index:
            classname += selected_background

        new_classnames.append(classname)
    return new_classnames


def create_checkerboard(height, width, size):
    """
    Create a checkerboard pattern.

    Args:
        height (int): The height of the checkerboard.
        width (int): The width of the checkerboard.
        size (int): The size of each square in the checkerboard.

    Returns:
        numpy.ndarray: The checkerboard pattern.
    """
    checkerboard = np.zeros((height, width), dtype=np.uint8)
    light_color = 200
    dark_color = 75
    for i in range(0, height, size):
        for j in range(0, width, size):
            color = light_color if (i // size + j // size) % 2 == 0 else dark_color
            checkerboard[i : i + size, j : j + size] = color

    # promote to 3 channels
    checkerboard = np.stack([checkerboard] * 3, axis=-1)

    return checkerboard


def xor_string(plaintext, secret):
    """
    XORs each byte of the plaintext with the corresponding byte of the secret.

    Args:
        plaintext (BytesIO object): The plaintext to be XORed.
        secret (bytes-like object): The secret used for XORing.

    Returns:
        bytes: The result of XORing the plaintext with the secret.
    """
    plaintext.seek(0)
    secret = io.BytesIO(secret)
    result = io.BytesIO()
    while True:
        plaintext_byte = plaintext.read(1)
        secret_byte = secret.read(1)
        if not plaintext_byte:
            break
        if not secret_byte:
            secret.seek(0)
            secret_byte = secret.read(1)
        result.write(bytes([plaintext_byte[0] ^ secret_byte[0]]))
    return result.getvalue()


def encode_string_with_nonce(plaintext, nonce):
    """
    Encodes a plaintext string using a nonce. This will be used for storing
    API keys and other sensitive information without making them readily
    available for credential scanning. This is not encryption. It's just masking
    the information.

    Args:
        plaintext (str): The plaintext string to be encoded.
        nonce (str): The nonce used for encoding.

    Returns:
        str: The encoded string.

    Example:
        >>> encode_string_with_nonce("Hello, World!", "unique-filename")
        'SGVsbG8sIFdvcmxkIQ=='
    """
    # convert plaintext to binary io
    plaintext = io.BytesIO(plaintext.encode("utf-8"))
    # hash the secret with sha256
    nonce = hashlib.sha256(nonce.encode("utf-8")).digest()

    result = xor_string(plaintext, nonce)
    return base64.b64encode(result).decode("utf-8")


def decode_string_with_nonce(ciphertext, nonce):
    # convert ciphertext to binary io
    try:
        ciphertext = io.BytesIO(base64.b64decode(ciphertext))
    except base64.binascii.Error:
        return None

    # hash the secret with sha256
    nonce = hashlib.sha256(nonce.encode("utf-8")).digest()

    result = xor_string(ciphertext, nonce)

    try:
        result = result.decode("utf-8")
    except UnicodeDecodeError:
        result = None

    return result



# (c) 2024 Niels Provos

import shutil
import threading
import json
import random
import string
import time
from typing import List
from PIL import Image
from io import BytesIO
from enum import Enum
import numpy as np
import cv2
from pathlib import Path
from .utils import (
    filename_previous_version,
    filename_add_version,
    apply_color_tint,
    create_checkerboard,
    encode_string_with_nonce,
    decode_string_with_nonce,
)
from .segmentation import mask_from_depth
from .upscaler import Upscaler
from .stabilityai import StabilityAI
from .camera import Camera
from .slice import ImageSlice


class CompositeMode(Enum):
    GRAYSCALE = 1
    CHECKERBOARD = 2
    NONE = 3


class AppState:
    __slots__ = (
        "_lock",
        "filename",
        "num_slices",
        "imgData",
        "imgThresholds",
        "depthMapData",
        "depth_model_name",
        "inpainting_model_name",
        "server_address",
        "api_key",
        "dark_mode",
        "image_slices",
        "selected_slice",
        "pipeline_spec",
        "depth_estimation_model",
        "segmentation_model",
        "selected_inpainting",
        "result_tinted",
        "grayscale_tinted",
        "checkerboard",
        "slice_pixel",
        "slice_pixel_depth",
        "slice_mask",
        "upscaler",
        "clipboard_image",
        "use_checkerboard",
        "multi_point_mode",
        "points_selected",
        "_camera",
        "_mesh_displacement",
    )
    SRV_DIR = "tmp-images"
    STATE_FILE = "appstate.json"
    IMAGE_FILE = "input_image.png"
    DEPTH_MAP_FILE = "depth_map.png"
    MAIN_IMAGE = "main_image.bmp"
    MODEL_FILE = "model.gltf"
    WORKFLOW = "workflow.json"

    cache = {}

    def __init__(self):
        # prevent concurrent writes
        self._lock = threading.Lock()

        self.filename = None
        self.num_slices = 5
        self.imgData = None  # PIL image
        self.imgThresholds = None
        self.depthMapData = None  # numpy array

        self.depth_model_name = None
        self.inpainting_model_name = None

        self.server_address = None
        self.api_key = None

        self.dark_mode = False

        # no JSON serialization for items below
        self.image_slices: List[ImageSlice] = []
        self.selected_slice = None
        self.pipeline_spec = None  # PipelineSpec() for inpainting
        self.depth_estimation_model = (
            None  # DepthEstimationModel() for depth estimation
        )
        self.segmentation_model = None  # SegmentationModel() for segmentation
        self.selected_inpainting = None
        self.result_tinted = None
        self.grayscale_tinted = None
        self.checkerboard = None
        self.slice_pixel = None
        self.slice_pixel_depth = None
        self.slice_mask = None
        self.upscaler = None
        self.clipboard_image = None

        self.use_checkerboard = False
        self.multi_point_mode = False
        self.points_selected = []

        self._camera = Camera(100.0, 500.0, 100.0)
        self._camera.camera_position = np.array([0.0, 0.0, -100.0], dtype=np.float32)

        self._mesh_displacement = 0.0

    @property
    def camera(self):
        return self._camera

    @property
    def mesh_displacement(self):
        return self._mesh_displacement

    @mesh_displacement.setter
    def mesh_displacement(self, value):
        if not isinstance(value, (float, int)) or value < 0:
            raise ValueError("mesh_displacement must be a non-negative number")
        self._mesh_displacement = value

    def camera_matrix(self):
        image_width, image_height = self.imgData.size
        return self._camera.camera_matrix(image_width, image_height)

    def get_cards(self):
        image_width, image_height = self.imgData.size
        cards = []
        for i, image_slice in enumerate(self.image_slices):
            card = image_slice.create_card(image_height, image_width, self.camera)
            cards.append(card)
        return cards

    def create_tints(self):
        """Precomputes the tings for visualizing slices."""
        # Apply the main color tint to the original image
        self.result_tinted = apply_color_tint(self.imgData, (150, 255, 0), 0.3)

        # Convert the image to grayscale and back to RGB
        grayscale = self.imgData.convert("L").convert("RGB")
        self.grayscale_tinted = apply_color_tint(grayscale, (0, 0, 150), 0.1)

    def apply_mask(self, image, mask):
        # Prepare masks; make sure they are the right size and mode
        if not isinstance(mask, Image.Image):
            mask = Image.fromarray(mask)
        mask = mask.convert("L")

        if image is not self.imgData:
            if not isinstance(image, Image.Image):
                image = Image.fromarray(image)
            image = image.convert("RGB")
            result_tinted = apply_color_tint(image, (150, 255, 0), 0.3)
            grayscale_tinted = apply_color_tint(
                image.convert("L").convert("RGB"), (0, 0, 150), 0.1
            )
        else:
            if self.result_tinted is None or self.grayscale_tinted is None:
                self.create_tints()
            result_tinted = self.result_tinted
            grayscale_tinted = self.grayscale_tinted

        # Combine the tinted and the grayscale image
        final_result = Image.composite(result_tinted, grayscale_tinted, mask)

        return final_result

    def change_slice_depth(self, slice_index, depth):
        """Changes the depth of the slice at the specified index."""
        assert slice_index >= 0 and slice_index < len(self.image_slices)

        if depth == self.image_slices[slice_index].depth:
            return slice_index

        # remove it from the lists
        image_slice = self.image_slices.pop(slice_index)
        image_slice.depth = depth

        return self.add_slice(image_slice)

    def add_slice(self, image_slice: ImageSlice):
        """Adds the image as a new slice at the provided depth."""
        if image_slice.filename is None:
            image_slice.filename = str(
                Path(self.filename) / f"image_slice_{len(self.image_slices)}.png"
            )
        # find the index where the depth should be inserted
        index = len(self.image_slices)
        for i, cur_slice in enumerate(self.image_slices):
            if image_slice.depth < cur_slice.depth:
                index = i
                break
        self.image_slices.insert(index, image_slice)

        if index > 0:
            # make sure the depth values are all unique
            for i in range(index, len(self.image_slices)):
                if self.image_slices[i].depth == self.image_slices[i - 1].depth:
                    self.image_slices[i].depth += 1

        return index

    def delete_slice(self, slice_index):
        """Deletes the slice at the specified index."""
        if slice_index < 0 or slice_index >= len(self.image_slices):
            return False
        self.image_slices.pop(slice_index)
        self.selected_slice = None
        self.slice_pixel = None
        self.slice_pixel_depth = None
        self.slice_mask = None

        return True

    def reset_image_slices(self):
        self.image_slices = []
        self.selected_slice = None
        self.selected_inpainting = None
        self.slice_pixel = None
        self.slice_mask = None
        self.slice_pixel_depth = None

    def balance_slices_depths(self):
        """Equally distribute the depths of the image slices."""
        depths = [
            int(i * 255 / (len(self.image_slices) - 1))
            for i in range(len(self.image_slices))
        ]
        for i in len(self.image_slices):
            self.image_slices[i].depth = depths[i]

    def depth_slice_from_pixel(self, pixel_x, pixel_y):
        depth = -1  # for log below
        if self.depthMapData is not None and self.imgThresholds is not None:
            depth = int(self.depthMapData[pixel_y, pixel_x])
            # find the depth that is bracketed by imgThresholds
            for i, threshold in enumerate(self.imgThresholds):
                if depth <= threshold:
                    threshold_min = int(self.imgThresholds[i - 1])
                    threshold_max = int(threshold)
                    break
            mask = mask_from_depth(self.depthMapData, threshold_min, threshold_max)

        return mask, depth

    def set_img_data(self, img_data):
        self.imgData = img_data.convert("RGB")
        self.depthMapData = None
        self.selected_slice = None
        self.selected_inpainting = None
        self.image_slices = []
        self.slice_pixel = None

        self.create_tints()

    def serve_model_file(self):
        """Serves the gltf model file."""
        model_path = Path(self.SRV_DIR) / Path(self.filename) / self.MODEL_FILE
        unique_id = int(time.time())
        return f"/{str(model_path)}?v={unique_id}"

    def serve_slice_image(self, slice_index):
        """Serves the image slice with the specified index."""
        assert slice_index >= 0 and slice_index < len(self.image_slices)
        image_path = self.checkerboard_filename(slice_index)
        if not image_path.exists():
            image = self.slice_image_composed(
                slice_index, mode=CompositeMode.CHECKERBOARD
            )
            image.save(image_path)
        image_path = Path(self.SRV_DIR) / image_path
        return f"/{str(image_path)}"

    def slice_image_composed(
        self, slice_index, mode: CompositeMode = CompositeMode.NONE
    ):
        """Composes the slice image over the main image."""
        assert slice_index >= 0 and slice_index < len(self.image_slices)
        slice_image = self.image_slices[slice_index].image
        if not isinstance(slice_image, Image.Image):
            slice_image = Image.fromarray(slice_image)

        composite = self.imgData
        if mode == CompositeMode.GRAYSCALE:
            if self.grayscale_tinted is None:
                self.create_tints()
            composite = self.grayscale_tinted
        elif mode == CompositeMode.CHECKERBOARD:
            if self.checkerboard is None:
                self.checkerboard = create_checkerboard(
                    slice_image.size[1], slice_image.size[0], 32
                )
            composite = Image.fromarray(self.checkerboard)

        full_image = Image.composite(
            slice_image, composite, slice_image.getchannel("A")
        )
        return full_image

    def serve_slice_image_composed(self, slice_index, mode: CompositeMode):
        """Serves the slice image composed over the gray main image."""
        full_image = self.slice_image_composed(slice_index, mode=mode)
        return self.serve_main_image(full_image)

    def serve_input_image(self):
        """Serves the input image from the state directory."""
        filename = Path(self.filename) / self.IMAGE_FILE
        if not filename.exists():
            if not Path(self.filename).exists():
                Path(self.filename).mkdir()
            self.imgData.save(filename, compress_level=1)
        filename = Path(self.SRV_DIR) / filename
        unique_id = int(time.time())
        return f"/{str(filename)}?v={unique_id}"

    def serve_main_image(self, image):
        """Serves the image using a temporary directory."""
        if not isinstance(image, Image.Image):
            image = Image.fromarray(image)
        output_dir = Path(self.filename)
        if not output_dir.exists():
            output_dir.mkdir()
        save_path = Path(self.filename) / self.MAIN_IMAGE
        image.save(save_path)
        image_path = Path(self.SRV_DIR) / save_path
        unique_id = int(time.time())
        return f"/{str(image_path)}?v={unique_id}"

    def workflow_path(self):
        """Returns the workflow path."""
        return Path(self.filename) / self.WORKFLOW

    def _make_filename(self, slice_index, suffix):
        assert slice_index >= 0 and slice_index < len(self.image_slices)
        file_path = Path(self.image_slices[slice_index].filename)
        return file_path.parent / f"{file_path.stem}_{suffix}.png"

    def checkerboard_filename(self, slice_index):
        """Returns the checkerboard filename for the specified slice index."""
        return self._make_filename(slice_index, "checkerboard")

    def depth_filename(self, slice_index):
        """Returns the depth filename for the specified slice index."""
        return self._make_filename(slice_index, "depth")

    def mask_filename(self, slice_index):
        """Returns the mask filename for the specified slice index."""
        return self._make_filename(slice_index, "mask")

    def upscaled_filename(self, slice_index):
        """Returns the upscaled filename for the specified slice index."""
        return self._make_filename(slice_index, "upscaled")

    def save_image_mask(self, slice_index, mask):
        """Saves the PIL image mask to a png file for the specified slice index."""
        mask_path = self.mask_filename(slice_index)
        mask.save(str(mask_path))

        return str(mask_path)

    def _read_image_slices(self):
        for i, image_slice in enumerate(self.image_slices):
            image_slice.read_image()
        print(f"Loaded {len(self.image_slices)} image slices")

    def save_image_slices(self, file_path):
        """
        Save the image slices to the specified file path.

        Args:
            file_path (str): The file path to save the image slices.
        """
        file_path = Path(file_path)
        for i, slice_image in enumerate(self.image_slices):
            if slice_image.filename is None:
                slice_image.filename = str(file_path / f"image_slice_{i}.png")
            slice_image.save_image()

    def _create_upscaler(self):
        if self.pipeline_spec is None:
            self.upscaler = Upscaler()
            return
        if self.inpainting_model_name == "stabilityai":
            model_name = "stabilityai"
            model = StabilityAI(self.api_key)
        else:
            model_name = "inpainting"
            model = self.pipeline_spec
            self.pipeline_spec.load_model()
        self.upscaler = Upscaler(model_name=model_name, external_model=model)

    def upscale_image(self, image, prompt="", negative_prompt=""):
        if self.upscaler is None:
            self._create_upscaler()
        upscaled_image = self.upscaler.upscale_image_tiled(
            image, overlap=64, prompt=prompt, negative_prompt=negative_prompt
        )
        return upscaled_image

    def upscale_slices(self):
        for i, slice_image in enumerate(self.image_slices):
            filename = self.upscaled_filename(i)
            if not Path(filename).exists():
                prompt = slice_image.positive_prompt
                negative_prompt = slice_image.negative_prompt
                print(
                    f"Upscaling image slice: {filename} with '{prompt}'/'{negative_prompt}'"
                )
                upscaled_image = self.upscale_image(
                    slice_image.image, prompt=prompt, negative_prompt=negative_prompt
                )
                upscaled_image.save(filename)
                print(f"Saved upscaled image slice: {filename}")

    def to_file(
        self,
        file_path,
        save_image_slices=True,
        save_depth_map=True,
        save_input_image=True,
    ):
        """
        Save the current state to a file.

        Args:
            file_path (str): The file path to save the state.
            save_image_slices (bool, optional): Whether to save the image slices. Defaults to True.
        """
        with self._lock:
            file_path = Path(file_path)
            # check if file path is a directory. if not create it
            if not file_path.is_dir():
                file_path.mkdir()

            # save the input image
            if save_input_image:
                img_file = file_path / AppState.IMAGE_FILE
                self.imgData.save(str(img_file))

            # save the depth map
            if self.depthMapData is not None and save_depth_map:
                depth_map_file = file_path / AppState.DEPTH_MAP_FILE
                image = Image.fromarray(self.depthMapData)
                image.save(str(depth_map_file))

            # save the image slices
            if save_image_slices:
                self.save_image_slices(file_path)

            state_file = file_path / AppState.STATE_FILE
            temp_file = state_file.with_suffix(".tmp")
            backup_file = state_file.with_suffix(".bak")

            try:
                # Write to temporary file
                with open(temp_file, "w", encoding="utf-8") as file:
                    file.write(self.to_json())

                # Create a backup of the current state file if it exists
                if state_file.exists():
                    shutil.move(state_file, backup_file)

                # Move the temporary file to the state file
                shutil.move(temp_file, state_file)

                # Optionally, remove the backup file if everything went fine
                if backup_file.exists():
                    backup_file.unlink()

            except Exception as e:
                # Clean up temporary and backup files if any error occurs
                if temp_file.exists():
                    temp_file.unlink()
                if backup_file.exists() and not state_file.exists():
                    # Restore from backup if the original state file is missing
                    shutil.move(backup_file, state_file)
                raise e  # Re-raise the error after cleanup

    @staticmethod
    def from_file(file_path):
        """
        Load the state from a file.

        Args:
            file_path (str): The file path to load the state from.

        Returns:
            AppState: The loaded state.
        """
        state_file = Path(file_path) / AppState.STATE_FILE
        with open(state_file, "r", encoding="utf-8") as file:
            state = AppState.from_json(file.read())

        state.fill_from_files(file_path)

        return state

    def fill_from_files(self, file_path):
        """
        This method reads the input image and depth map files from the specified file path and updates the state
        with the corresponding image data and depth map data.

        Args:
            file_path (str): The file path to load the state from.

        Returns:
            None
        """
        # read the input image
        img_file = Path(file_path) / AppState.IMAGE_FILE
        self.imgData = Image.open(img_file)
        self._read_image_slices()

        # read the depth map and turn it into a numpy array
        depth_map_file = Path(file_path) / AppState.DEPTH_MAP_FILE
        if depth_map_file.exists():
            self.depthMapData = np.array(Image.open(depth_map_file))

    @staticmethod
    def from_cache(file_path):
        """
        Retrieve the state from the cache or load it from a file.

        Args:
            file_path (str): The file path to load the state from.

        Returns:
            AppState: The loaded state.
        """
        if file_path in AppState.cache:
            return AppState.cache[file_path]
        state = AppState.from_file(file_path)
        AppState.cache[file_path] = state
        return state

    @staticmethod
    def from_file_or_new(file_path):
        """
        Load the state from a file or create a new state if the file path is None.

        Args:
            file_path (str): The file path to load the state from.

        Returns:
            tuple: A tuple containing the loaded state and the file path.
        """
        if file_path is None:
            state = AppState()
            file_path = AppState.random_filename()
            state.filename = file_path
        else:
            state = AppState.from_file(file_path)
        AppState.cache[file_path] = state
        return state, file_path

    @staticmethod
    def random_filename():
        """
        Generate a random filename.

        Returns:
            str: The random filename.
        """
        return "appstate-" + "".join(
            random.choices(string.ascii_letters + string.digits, k=8)
        )

    def to_json(self):
        """
        Convert the state to a JSON string.

        Returns:
            str: The JSON string representation of the state.
        """

        image_depths = [image_slice.depth for image_slice in self.image_slices]
        image_slices_filenames = [
            image_slice.filename for image_slice in self.image_slices
        ]
        positive_prompts = [
            image_slice.positive_prompt for image_slice in self.image_slices
        ]
        negative_prompts = [
            image_slice.negative_prompt for image_slice in self.image_slices
        ]

        data = {
            "filename": self.filename,
            "num_slices": self.num_slices,
            "imgThresholds": self.imgThresholds,
            "image_depths": image_depths,
            "image_slices_filenames": image_slices_filenames,
            "positive_prompts": positive_prompts,
            "negative_prompts": negative_prompts,
            "dark_mode": self.dark_mode,
            "mesh_displacement": self._mesh_displacement,
        }

        # merge the camera data
        camera_dict = self._camera.to_json()
        data.update(camera_dict)

        if self.depth_model_name is not None:
            data["depth_model_name"] = self.depth_model_name
        if self.inpainting_model_name is not None:
            data["inpainting_model_name"] = self.inpainting_model_name
        if self.server_address is not None:
            data["server_address"] = self.server_address
        if self.api_key is not None:
            data["api_key"] = encode_string_with_nonce(self.api_key, self.filename)
        return json.dumps(data)

    @staticmethod
    def from_json(json_data):
        """
        Load the state from a JSON string.

        Args:
            json_data (str): The JSON string representation of the state.

        Returns:
            AppState: The loaded state.
        """
        state = AppState()
        if json_data is None:
            return state

        data = json.loads(json_data)

        state.filename = data["filename"]
        state.num_slices = data["num_slices"] if "num_slices" in data else 5
        state.imgThresholds = data["imgThresholds"]

        image_depths = (
            data["image_depths"] if "image_depths" in data else state.imgThresholds[1:]
        )
        image_slices_filenames = data["image_slices_filenames"]

        state.depth_model_name = (
            data["depth_model_name"] if "depth_model_name" in data else None
        )
        state.inpainting_model_name = (
            data["inpainting_model_name"] if "inpainting_model_name" in data else None
        )

        empty = [""] * len(image_slices_filenames)
        positive_prompts = (
            data["positive_prompts"] if "positive_prompts" in data else empty
        )
        negative_prompts = (
            data["negative_prompts"] if "negative_prompts" in data else empty
        )

        # check dats structures have consistent lengths
        assert len(image_slices_filenames) == len(image_depths)
        assert len(image_slices_filenames) == len(positive_prompts)
        assert len(image_slices_filenames) == len(negative_prompts)

        state.image_slices = [
            ImageSlice(
                depth=depth,
                filename=filename,
                positive_prompt=positive_prompt,
                negative_prompt=negative_prompt,
            )
            for depth, filename, positive_prompt, negative_prompt in zip(
                image_depths, image_slices_filenames, positive_prompts, negative_prompts
            )
        ]

        state.server_address = (
            data["server_address"] if "server_address" in data else None
        )
        state.api_key = (
            decode_string_with_nonce(data["api_key"], state.filename)
            if "api_key" in data
            else None
        )

        state.dark_mode = data["dark_mode"] if "dark_mode" in data else False

        if "mesh_displacement" in data:
            state.mesh_displacement = data["mesh_displacement"]

        state._camera = Camera.from_json(data)

        # check that all filenames start with the filename as prefix
        state.check_pathnames()

        return state

    def check_pathnames(self):
        """Check that all pathnames are valid."""

        cwd = Path().cwd()
        root = Path(self.filename).resolve()
        assert str(root).startswith(
            str(cwd / "appstate-")
        ), f"Invalid filename: {self.filename}"

        for image_slice in self.image_slices:
            filename = Path(image_slice.filename).resolve()
            assert str(filename).startswith(str(root)), f"Invalid filename: {filename}"



# (c) 2024 Niels Provos

import numpy as np


class Camera:
    __slots__ = (
        "_camera_position",
        "_camera_distance",
        "_max_distance",
        "_focal_length",
        "_sensor_width",
    )

    def __init__(
        self, distance=100, max_distance=500, focal_length=100, sensor_width=35.0
    ):
        """
        Initializes a Camera object.

        It isn't really a camera but encapsulates related properties.

        Args:
            distance (float): The distance of the camera from the scene.
            max_distance (float): The maximum distance/depth of the scene.
            focal_length (float): The focal length of the camera.
            sensor_width (float, optional): The width of the camera sensor. Defaults to 35.0.

        Returns:
            None
        """
        self.camera_position = np.array([0, 0, -distance], dtype=np.float32)
        self.camera_distance = distance
        self.max_distance = max_distance
        self.focal_length = focal_length
        self.sensor_width = sensor_width

    def focal_length_px(self, image_width):
        """
        Calculate the focal length in pixels.

        Args:
            image_width (int): The width of the image.

        Returns:
            float: The focal length in pixels.
        """
        return (image_width * self.focal_length) / self.sensor_width

    def camera_matrix(self, image_width, image_height):
        """
        Calculate the camera matrix.

        Args:
            image_width (int): The width of the image.
            image_height (int): The height of the image.

        Returns:
            np.ndarray: The camera matrix.
        """

        fl_px = self.focal_length_px(image_width)
        return np.array(
            [[fl_px, 0, image_width / 2], [0, fl_px, image_height / 2], [0, 0, 1]],
            dtype=np.float32,
        )

    def to_json(self):
        return {
            "position": self._camera_position.tolist(),
            "camera_distance": self._camera_distance,
            "max_distance": self._max_distance,
            "focal_length": self._focal_length,
        }

    @staticmethod
    def from_json(data):
        camera = Camera()
        if "position" in data:
            camera.camera_position = np.array(data["position"], dtype=np.float32)
        if "camera_distance" in data:
            camera.camera_distance = data["camera_distance"]
        if "max_distance" in data:
            camera.max_distance = data["max_distance"]
        if "focal_length" in data:
            camera.focal_length = data["focal_length"]

        return camera

    def __str__(self):
        return (
            f"Camera(distance={self._camera_distance}, max_distance={self._max_distance}, "
            f"focal_length={self._focal_length}, position={self._camera_position})"
        )

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, Camera):
            return False
        return (
            self.camera_distance == other.camera_distance
            and self.max_distance == other.max_distance
            and self.focal_length == other.focal_length
        )

    @property
    def camera_position(self):
        return self._camera_position

    @camera_position.setter
    def camera_position(self, value):
        if (
            not isinstance(value, np.ndarray)
            or value.dtype != np.float32
            or value.shape != (3,)
        ):
            raise ValueError(
                "camera_position must be a numpy array of dtype 'float32' with shape (3,)"
            )
        self._camera_position = value

    @property
    def camera_distance(self):
        return self._camera_distance

    @camera_distance.setter
    def camera_distance(self, value):
        if not isinstance(value, (float, int)) or value < 0:
            raise ValueError("camera_distance must be a non-negative number")
        self._camera_distance = value

    @property
    def max_distance(self):
        return self._max_distance

    @max_distance.setter
    def max_distance(self, value):
        if not isinstance(value, (float, int)) or value < 0:
            raise ValueError("max_distance must be a non-negative number")
        self._max_distance = value

    @property
    def focal_length(self):
        return self._focal_length

    @focal_length.setter
    def focal_length(self, value):
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError("focal_length must be a positive number")
        self._focal_length = value

    @property
    def sensor_width(self):
        return self._sensor_width

    @sensor_width.setter
    def sensor_width(self, value):
        if not isinstance(value, (float, int)) or value <= 0:
            raise ValueError("sensor_width must be a positive number")
        self._sensor_width = value



import unittest
from PIL import Image
from .utils import (
    find_bounding_box,
    find_square_from_bounding_box,
    find_square_bounding_box,
    filename_add_version,
    filename_previous_version,
    highlight_selected_element,
    encode_string_with_nonce,
    decode_string_with_nonce,
)


class TestFindBoundingBox(unittest.TestCase):
    def test_find_bounding_box(self):
        # Create a mask image with a bounding box
        mask_image = Image.new("L", (100, 100))
        mask_image.paste(255, (20, 30, 80, 70))

        # Call the function
        result = find_bounding_box(mask_image, padding=0)

        # Define the expected result
        expected_result = (20, 30, 79, 69)

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_find_bounding_box_with_padding(self):
        # Create a mask image with a bounding box
        mask_image = Image.new("L", (100, 100))
        mask_image.paste(255, (20, 30, 80, 70))

        # Call the function with padding
        result = find_bounding_box(mask_image, padding=10)

        # Define the expected result
        expected_result = (10, 20, 89, 79)

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_find_bounding_box_with_empty_mask(self):
        # Create an empty mask image
        mask_image = Image.new("L", (100, 100))

        # Call the function
        result = find_bounding_box(mask_image)

        # Define the expected result
        expected_result = (0, 0, 0, 0)

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_find_square_from_bounding_box(self):
        # Define the bounding box coordinates
        xmin, ymin, xmax, ymax = 20, 30, 80, 70

        # Call the function
        result = find_square_from_bounding_box(xmin, ymin, xmax, ymax)

        # Define the expected result
        expected_result = (20, 20, 80, 80)

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_find_square_bounding_box(self):
        # Create an empty mask image
        mask_image = Image.new("L", (100, 100))
        mask_image.paste(255, (20, 30, 80, 70))

        result = find_square_bounding_box(mask_image, padding=10)

        expected_result = (11, 11, 89, 89)

        self.assertEqual(result, expected_result)

    def test_find_square_bounding_box_outside(self):
        # Create an empty mask image
        mask_image = Image.new("L", (100, 100))
        mask_image.paste(255, (20, 30, 80, 70))

        result = find_square_bounding_box(mask_image, padding=30)

        expected_result = (0, 0, 100, 100)

        self.assertEqual(result, expected_result)


class TestFilenameAddVersion(unittest.TestCase):
    def test_filename_add_version(self):
        # Define the input filename
        filename = "/path/to/image.png"

        # Call the function
        result = filename_add_version(filename)

        # Define the expected result
        expected_result = "/path/to/image_v2.png"

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

        # Call the function
        result = filename_add_version(result)

        # Define the expected result
        expected_result = "/path/to/image_v3.png"

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_filename_previous_version(self):
        # Define the input filename
        filename = "/path/to/image_v3.png"

        # Call the function
        result = filename_previous_version(filename)

        # Define the expected result
        expected_result = "/path/to/image_v2.png"

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

        # Call the function
        result = filename_previous_version(result)

        # Define the expected result
        expected_result = "/path/to/image.png"

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

        # Call the function with a filename without version
        filename = "/path/to/image.png"
        result = filename_previous_version(filename)

        # Define the expected result
        expected_result = None

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)


class TestHighLightSelectedElement(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.classnames = [
            "text-white",
            "text-brown",
            "text-black",
        ]
        self.highlight_classname = "color-is-selected"

    def test_highlight_selected_element(self):
        result = highlight_selected_element(
            self.classnames, 0, self.highlight_classname
        )

        expected_result = self.classnames.copy()
        expected_result[0] += f" {self.highlight_classname}"

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_highlight_selected_element_remove(self):
        classnames = self.classnames.copy()
        expected_result = self.classnames
        classnames[1] += f" {self.highlight_classname}"

        result = highlight_selected_element(classnames, None, self.highlight_classname)

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)


class TestEncodeStringWithNonce(unittest.TestCase):
    def test_encode_string_with_nonce(self):
        # Define the plaintext and nonce
        plaintext = "Hello, World!"
        nonce = "unique-filename"

        # Call the function
        result = encode_string_with_nonce(plaintext, nonce)

        self.assertNotEqual(result, plaintext)

        decoded = decode_string_with_nonce(result, nonce)

        self.assertEqual(decoded, plaintext)

        decoded = decode_string_with_nonce(result, "wrong-nonce")

        self.assertNotEqual(decoded, plaintext)

    def test_decode_with_string_bad_data(self):
        nonce = "unique-filename"
        input = "bad-data"

        decoded = decode_string_with_nonce(input, nonce)

        self.assertEqual(decoded, None)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch
from PIL import Image
from .automatic1111 import create_img2img_payload, make_models_request


class TestCreateImg2ImgPayload(unittest.TestCase):
    def setUp(self):
        self.input_image = Image.new("RGB", (100, 100))
        self.positive_prompt = "Generate a beautiful image"
        self.negative_prompt = "Avoid using bright colors"
        self.mask_image = Image.new("L", (100, 100))
        self.strength = 0.8
        self.steps = 20
        self.cfg_scale = 8.0

    def test_create_img2img_payload_without_mask(self):
        with patch(
            "parallax_maker.automatic1111.to_image_url",
            return_value="data:image/png;base64,",
        ) as mock_to_image_url:
            payload = create_img2img_payload(
                self.input_image,
                self.positive_prompt,
                self.negative_prompt,
                mask_image=None,
                strength=self.strength,
                steps=self.steps,
                cfg_scale=self.cfg_scale,
            )

        self.assertEqual(payload["init_images"], ["data:image/png;base64,"])
        self.assertEqual(payload["prompt"], self.positive_prompt)
        self.assertEqual(payload["negative_prompt"], self.negative_prompt)
        self.assertEqual(payload["denoising_strength"], self.strength)
        self.assertEqual(payload["width"], 100)
        self.assertEqual(payload["height"], 100)
        self.assertEqual(payload["steps"], self.steps)
        self.assertEqual(payload["cfg_scale"], self.cfg_scale)
        self.assertNotIn("mask", payload)

    def test_create_img2img_payload_with_mask(self):
        with patch(
            "parallax_maker.automatic1111.to_image_url",
            return_value="data:image/png;base64,",
        ) as mock_to_image_url:
            payload = create_img2img_payload(
                self.input_image,
                self.positive_prompt,
                self.negative_prompt,
                mask_image=self.mask_image,
                strength=self.strength,
                steps=self.steps,
                cfg_scale=self.cfg_scale,
            )

        self.assertEqual(payload["init_images"], ["data:image/png;base64,"])
        self.assertEqual(payload["prompt"], self.positive_prompt)
        self.assertEqual(payload["negative_prompt"], self.negative_prompt)
        self.assertEqual(payload["denoising_strength"], self.strength)
        self.assertEqual(payload["width"], 100)
        self.assertEqual(payload["height"], 100)
        self.assertEqual(payload["steps"], self.steps)
        self.assertEqual(payload["cfg_scale"], self.cfg_scale)
        self.assertEqual(payload["mask"], "data:image/png;base64,")


class TestMakeModelsRequest(unittest.TestCase):
    @patch("parallax_maker.automatic1111.requests.get")
    def test_make_models_request(self, mock_get):
        mock_response = [
            {
                "title": "sd_xl_base_1.0.safetensors [31e35c80fc]",
                "model_name": "sd_xl_base_1.0",
                "hash": "31e35c80fc",
                "sha256": "31e35c80fc4829d14f90153f4c74cd59c90b779f6afe05a74cd6120b893f7e5b",
                "filename": "models/Stable-diffusion/sd_xl_base_1.0.safetensors",
                "config": None,
            },
            {
                "title": "sd_xl_refiner_1.0.safetensors [7440042bbd]",
                "model_name": "sd_xl_refiner_1.0",
                "hash": "7440042bbd",
                "sha256": "7440042bbdc8a24813002c09b6b69b64dc90fded4472613437b7f55f9b7d9c5f",
                "filename": "models/Stable-diffusion/sd_xl_refiner_1.0.safetensors",
                "config": None,
            },
            {
                "title": "v2-1_768-ema-pruned.ckpt [ad2a33c361]",
                "model_name": "v2-1_768-ema-pruned",
                "hash": "ad2a33c361",
                "sha256": "ad2a33c361c1f593c4a1fb32ea81afce2b5bb7d1983c6b94793a26a3b54b08a0",
                "filename": "models/Stable-diffusion/v2-1_768-ema-pruned.ckpt",
                "config": None,
            },
        ]

        mock_get.return_value.json.return_value = mock_response

        server_address = "localhost:7860"
        models = make_models_request(server_address)

        self.assertEqual(
            models, ["sd_xl_base_1.0", "sd_xl_refiner_1.0", "v2-1_768-ema-pruned"]
        )
        mock_get.assert_called_once_with(
            url="http://localhost:7860/sdapi/v1/sd-models", timeout=3
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# (c) 2024 Niels Provos
#

import argparse
import os
from pathlib import Path

from PIL import Image

from .controller import AppState
from .depth import DepthEstimationModel
from .segmentation import generate_depth_map
from .utils import postprocess_depth_map
from .webui import export_state_as_gltf


def compute_depth_map_for_slices(state: AppState, postprocess: bool = True):
    depth_maps = []

    model_name = state.depth_model_name if state.depth_model_name else "midas"
    model = DepthEstimationModel(model=model_name)
    for i, image_slice in enumerate(state.image_slices):
        filename = image_slice.filename
        print(f"Processing {filename}")

        image = image_slice.image

        depth_map = generate_depth_map(image[:, :, :3], model)

        tmp_filename = state._make_filename(i, "depth_tmp")
        depth_image = Image.fromarray(depth_map)
        depth_image.save(tmp_filename, compress_level=9)

        if postprocess:
            image_alpha = image[:, :, 3]
            depth_map = postprocess_depth_map(depth_map, image_alpha, final_blur=50)

        depth_image = Image.fromarray(depth_map)

        output_filename = Path(state.filename) / (Path(filename).stem + "_depth.png")

        depth_image.save(output_filename, compress_level=1)
        print(f"Saved depth map to {output_filename}")

        depth_maps.append(output_filename)
    return depth_maps


def main():
    os.environ["DISABLE_TELEMETRY"] = "YES"
    os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

    # get arguments from the command line
    # -i name of the state file
    # -o output for the gltf file
    parser = argparse.ArgumentParser(
        description="Create a glTF file from the state file"
    )
    parser.add_argument("-i", "--state_file", type=str, help="Path to the state file")
    parser.add_argument(
        "-o",
        "--output_path",
        type=str,
        default="output",
        help="Path to save the glTF file",
    )
    parser.add_argument(
        "-n",
        "--no-inline",
        action="store_true",
        help="Do not inline images in the glTF file",
    )
    parser.add_argument(
        "-d", "--depth", action="store_true", help="Compute depth maps for slices"
    )
    parser.add_argument(
        "-s", "--scale", type=float, default=0.0, help="Displacement scale factor"
    )
    args = parser.parse_args()

    state = AppState.from_file(args.state_file)

    output_path = Path(args.output_path)
    if not output_path.exists():
        output_path.mkdir(parents=True)

    if args.depth:
        compute_depth_map_for_slices(state)

    gltf_path = export_state_as_gltf(
        state,
        args.output_path,
        state.camera,
        displacement_scale=args.scale,
        inline_images=not args.no_inline,
        support_dof=True,
    )
    print(f"Exported glTF to {gltf_path}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# (c) 2024 Niels Provos
#
"""
Upscale Image Tiling

This module provides functionality to upscale an image using a pre-trained deep learning model,
specifically Swin2SR for Super Resolution. The upscaling process breaks the input image into smaller
tiles.  Each tile is individually upscaled using the model, and then these upscaled tiles are reassembled
into the final image. The tile-based approach helps manage memory usage and can handle larger
images by processing small parts one at a time.

For better integration of the tiles into the resultant upscaled image, tiles are overlapped and
blended to ensure smoother transitions between tiles.
"""

from PIL import Image
import torch
from transformers import AutoImageProcessor, Swin2SRForImageSuperResolution

import numpy as np
from scipy.ndimage import zoom

from .utils import torch_get_device, premultiply_alpha_numpy, find_bounding_box
import argparse


class Upscaler:
    def __init__(self, model_name="swin2sr", external_model=None):
        assert model_name in ["swin2sr", "simple", "inpainting", "stabilityai"]
        self.model_name = model_name
        self.model = None
        self.image_processor = None
        self.external_model = external_model
        self.tile_size = 512
        self.scale_factor = 2

    def create_model(self):
        if self.model_name == "swin2sr":
            self.model, self.image_processor = self.load_swin2sr_model()
        elif self.model_name == "simple":
            self.model, self.image_processor = None, None
            self.tile_size = 512
        elif self.model_name == "inpainting":
            assert self.external_model is not None
            assert self.external_model.get_dimension() is not None
            self.model, self.image_processor = self.external_model, None
            self.tile_size = self.external_model.get_dimension() // 2
        elif self.model_name == "stabilityai":
            self.model, self.image_processor = self.external_model, None
            self.tile_size = 1024
            self.scale_factor = 4

    def __eq__(self, other):
        if not isinstance(other, Upscaler):
            return False
        return self.model_name == other.model_name

    def load_swin2sr_model(self):
        # Initialize the image processor and model
        image_processor = AutoImageProcessor.from_pretrained(
            "caidas/swin2SR-classical-sr-x2-64"
        )
        model = Swin2SRForImageSuperResolution.from_pretrained(
            "caidas/swin2SR-classical-sr-x2-64"
        )
        model.to(torch_get_device())

        return model, image_processor

    def upscale_image_tiled(self, image, overlap=64, prompt=None, negative_prompt=None):
        """
        Upscales an image using a tiled approach.

        Args:
            image (PIL.Image or np.ndarray): The input image array.
            overlap (int, optional): The overlap between adjacent tiles. Defaults to 64.

        Returns:
            np.ndarray: The upscaled image array.
        """
        if isinstance(image, Image.Image):
            image = np.array(image)

        # initializes parameters we need below
        if self.model is None:
            self.create_model()

        alpha = None
        bounding_box = None
        if image.shape[2] == 4:  # RGBA image
            alpha = image[:, :, 3]
            bounding_box = find_bounding_box(alpha, padding=0)

            # scale the size of the alpha channel using bicubic interpolation
            alpha = zoom(alpha, self.scale_factor, order=3)
            image = image[:, :, :3]  # Remove alpha channel

            # crop the image to the bounding box
            orig_image = image
            image = image[
                bounding_box[1] : bounding_box[3], bounding_box[0] : bounding_box[2]
            ]

        # Ensure the overlap can be divided by the scale factor
        if overlap % self.scale_factor != 0:
            overlap = (overlap // self.scale_factor + 1) * self.scale_factor

        # Calculate the number of tiles
        height, width, _ = image.shape
        step_size = self.tile_size - overlap // self.scale_factor
        num_tiles_x = (width + step_size - 1) // step_size
        num_tiles_y = (height + step_size - 1) // step_size

        # Create a new array to store the upscaled result
        upscaled_height = height * self.scale_factor
        upscaled_width = width * self.scale_factor
        upscaled_image = np.zeros((upscaled_height, upscaled_width, 3), dtype=np.uint8)

        # Iterate over the tiles
        for y in range(num_tiles_y):
            for x in range(num_tiles_x):
                # Calculate the coordinates of the current tile
                left = x * step_size
                top = y * step_size
                right = min(left + self.tile_size, width)
                bottom = min(top + self.tile_size, height)

                # make sure we process a full tile
                if x > 0 and right - left < self.tile_size:
                    left = right - self.tile_size
                    assert left >= 0
                if y > 0 and bottom - top < self.tile_size:
                    top = bottom - self.tile_size
                    assert top >= 0

                print(
                    f"Processing tile ({y}, {x}) with coordinates ({left}, {top}, {right}, {bottom})"
                )

                # Extract the current tile from the image
                tile = image[top:bottom, left:right]
                # XXX - revisit whether we can keep this as a numpy array
                tile = Image.fromarray(tile)

                cur_width, cur_height = tile.size
                if cur_width % 64 != 0 or cur_height % 64 != 0:
                    new_width = (
                        cur_width + (64 - cur_width % 64)
                        if cur_width % 64 != 0
                        else cur_width
                    )
                    new_height = (
                        cur_height + (64 - cur_height % 64)
                        if cur_height % 64 != 0
                        else cur_height
                    )
                    tile = tile.resize((new_width, new_height))
                upscaled_tile = self.upscale_tile(tile, prompt, negative_prompt)
                if tile.size != (cur_width, cur_height):
                    upscaled_tile = upscaled_tile.resize(
                        (cur_width * self.scale_factor, cur_height * self.scale_factor)
                    )
                upscaled_tile = np.array(upscaled_tile)

                # Calculate the coordinates to paste the upscaled tile
                place_left = left * self.scale_factor
                place_top = top * self.scale_factor
                place_right = place_left + upscaled_tile.shape[1]
                place_bottom = place_top + upscaled_tile.shape[0]

                self.integrate_tile(
                    upscaled_tile,
                    upscaled_image,
                    place_left,
                    place_top,
                    place_right,
                    place_bottom,
                    x,
                    y,
                    overlap,
                )

        # Combine the upscaled image with the alpha channel if present
        if alpha is not None:
            image = zoom(orig_image, (self.scale_factor, self.scale_factor, 1), order=3)
            bounding_box = [coord * self.scale_factor for coord in bounding_box]
            image[
                bounding_box[1] : bounding_box[3], bounding_box[0] : bounding_box[2]
            ] = upscaled_image
            upscaled_image = np.dstack((image, alpha))
            upscaled_image = premultiply_alpha_numpy(upscaled_image)
        else:
            upscaled_image = Image.fromarray(upscaled_image)

        return upscaled_image

    @staticmethod
    def integrate_tile(tile, image, left, top, right, bottom, tile_x, tile_y, overlap):
        height, width, _ = tile.shape

        # xxx - should be move before upscaling
        if overlap >= height or overlap >= width:
            return

        # Create an alpha channel for the tile
        alpha = np.ones((height, width), dtype=np.float32)
        if tile_x > 0 and tile_y > 0:
            alpha[:overlap, :overlap] = np.outer(
                np.linspace(0, 1, overlap), np.linspace(0, 1, overlap)
            )
        if tile_x > 0:
            new_alpha = np.tile(np.linspace(0, 1, overlap), (height, 1))
            alpha[:, :overlap] = np.minimum(alpha[:, :overlap], new_alpha)

        if tile_y > 0:
            new_alpha = np.tile(np.linspace(0, 1, overlap).reshape(-1, 1), (1, width))
            alpha[:overlap, :] = np.minimum(alpha[:overlap, :], new_alpha)

        # Reshape the alpha channel to match the tile shape
        alpha = alpha.reshape(height, width, 1)

        # Apply the tile with alpha blending to the image
        image[top:bottom, left:right] = (
            alpha * tile + (1 - alpha) * image[top:bottom, left:right]
        ).astype(np.uint8)

    def upscale_tile(self, tile, prompt=None, negative_prompt=None):
        if self.model_name == "swin2sr":
            return self._upscale_tile_swin2sr(tile)
        elif self.model_name == "simple":
            return self._upscale_tile_simple(tile)
        elif self.model_name == "inpainting":
            return self._upscale_tile_inpainting(tile, prompt, negative_prompt)
        elif self.model_name == "stabilityai":
            return self._upscale_tile_stabilityai(tile, prompt, negative_prompt)

    def _upscale_tile_stabilityai(self, tile, prompt, negative_prompt):
        upscaled_image = self.external_model.upscale_image(
            tile, prompt, negative_prompt=negative_prompt
        )
        width, height = tile.size
        upscaled_image = upscaled_image.resize(
            (width * self.scale_factor, height * self.scale_factor), Image.LANCZOS
        )
        return upscaled_image

    def _upscale_tile_inpainting(self, tile, prompt, negative_prompt):
        rescaled_tile = tile.resize((tile.size[0] * 2, tile.size[1] * 2), Image.LANCZOS)
        rescaled_tile = np.array(rescaled_tile)
        mask_image = np.ones(rescaled_tile.shape[:2], dtype=np.uint8)
        mask_image *= 255
        scale = 2 if len(prompt) or len(negative_prompt) else 0
        tile = self.external_model.inpaint(
            prompt,
            negative_prompt,
            rescaled_tile,
            mask_image,
            strength=0.25,
            guidance_scale=scale,
            num_inference_steps=75,
            padding=0,
            blur_radius=0,
        )
        tile = tile.convert("RGB")
        return tile

    def _upscale_tile_simple(self, tile):
        """
        Upscales a tile using a simple bicubic interpolation.

        Args:
            tile: The input tile to be upscaled.

        Returns:
            An Image object representing the upscaled tile.
        """
        return tile.resize((tile.width * 2, tile.height * 2), Image.BICUBIC)

    def _upscale_tile_swin2sr(self, tile):
        """
        Upscales a tile using a given model and image processor.

        Args:
            model: The model used for upscaling the tile.
            image_processor: The image processor used to preprocess the tile.
            tile: The input tile to be upscaled.

        Returns:
            An Image object representing the upscaled tile.
        """
        inputs = self.image_processor(tile, return_tensors="pt")
        inputs = {name: tensor.to(self.model.device) for name, tensor in inputs.items()}

        with torch.no_grad():
            outputs = self.model(**inputs)

        output = outputs.reconstruction.squeeze().float().cpu().clamp_(0, 1).numpy()
        output = np.moveaxis(output, source=0, destination=-1)
        output = (output * 255.0).round().astype(np.uint8)

        return Image.fromarray(output)


if __name__ == "__main__":
    from inpainting import InpaintingModel
    from stabilityai import StabilityAI

    parser = argparse.ArgumentParser(description="Image Upscaling")
    parser.add_argument(
        "-i", "--input", type=str, default="input.jpg", help="Input image path"
    )
    parser.add_argument(
        "-p", "--prompt", type=str, default="a sci-fi robot in a futuristic laboratory"
    )
    parser.add_argument("--api-key", type=str)
    args = parser.parse_args()

    if args.api_key:
        model_name = "stabilityai"
        model = StabilityAI(args.api_key)
    else:
        model_name = "inpainting"
        model = InpaintingModel()
        model.load_model()
    upscaler = Upscaler(model_name=model_name, external_model=model)
    image = Image.open(args.input)
    upscaled_image = upscaler.upscale_image_tiled(
        image,
        overlap=64,
        prompt=args.prompt,
    )
    upscaled_image.save("upscaled_image.png")



import unittest
from unittest.mock import patch, Mock
from PIL import Image
from io import BytesIO
import requests
from .stabilityai import StabilityAI


class TestStabilityAI(unittest.TestCase):

    @patch("parallax_maker.stabilityai.requests.get")
    def test_validate_key_success(self, mock_get):
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"credits": 100}
        mock_get.return_value = mock_response

        ai = StabilityAI(api_key="fake_api_key")
        success, credits = ai.validate_key()

        self.assertTrue(success)
        self.assertEqual(credits, 100)

    @patch("parallax_maker.stabilityai.requests.get")
    def test_validate_key_failure(self, mock_get):
        mock_response = Mock()
        mock_response.status_code = 401
        mock_get.return_value = mock_response

        ai = StabilityAI(api_key="fake_api_key")
        success, _ = ai.validate_key()

        self.assertFalse(success)

    @patch("parallax_maker.stabilityai.requests.post")
    def test_generate_image_success(self, mock_post):
        # Prepare mocked response
        image = Image.new("RGB", (100, 100))
        image_data = BytesIO()
        image.save(image_data, format="PNG")
        image_data.seek(0)

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = image_data.read()
        mock_post.return_value = mock_response

        ai = StabilityAI(api_key="fake_api_key")
        result_image = ai.generate_image(prompt="A dog running in the park")

        self.assertIsInstance(result_image, Image.Image)
        self.assertEqual(result_image.size, (100, 100))

    @patch("parallax_maker.stabilityai.requests.post")
    def test_generate_image_failure(self, mock_post):
        mock_response = Mock()
        mock_response.status_code = 400
        mock_response.json.return_value = {"error": "Bad request"}
        mock_post.return_value = mock_response

        ai = StabilityAI(api_key="fake_api_key")
        with self.assertRaises(Exception):
            ai.generate_image(prompt="A dog running in the park")

    @patch("parallax_maker.stabilityai.requests.post")
    def test_image_to_image_success(self, mock_post):
        image = Image.new("RGB", (100, 100))
        input_image = Image.new("RGB", (100, 100))
        image_data = BytesIO()
        image.save(image_data, format="PNG")
        image_data.seek(0)

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = image_data.read()
        mock_post.return_value = mock_response

        ai = StabilityAI(api_key="fake_api_key")
        result_image = ai.image_to_image(input_image, prompt="A cat on a sofa")

        self.assertIsInstance(result_image, Image.Image)
        self.assertEqual(result_image.size, (100, 100))

    @patch("parallax_maker.stabilityai.requests.post")
    def test_image_to_image_invalid_size(self, mock_post):
        small_image = Image.new("RGB", (50, 50))
        ai = StabilityAI(api_key="fake_api_key")

        with self.assertRaises(ValueError):
            ai.image_to_image(small_image, prompt="A cat on a sofa")

    @patch("parallax_maker.stabilityai.requests.post")
    def test_inpaint_image_success(self, mock_post):
        image = Image.new("RGB", (100, 100))
        mask = Image.new("1", (100, 100))
        output_image = Image.new("RGB", (100, 100))
        image_data = BytesIO()
        output_image.save(image_data, format="PNG")
        image_data.seek(0)

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = image_data.read()
        mock_post.return_value = mock_response

        ai = StabilityAI(api_key="fake_api_key")
        result_image = ai.inpaint_image(
            image, mask, prompt="A beach sunset", strength=0.5
        )

        self.assertIsInstance(result_image, Image.Image)
        self.assertEqual(result_image.size, (100, 100))

    @patch("parallax_maker.stabilityai.requests.post")
    def test_inpaint_image_invalid_size(self, mock_post):
        small_image = Image.new("RGB", (50, 50))
        mask = Image.new("1", (50, 50))
        ai = StabilityAI(api_key="fake_api_key")

        with self.assertRaises(ValueError):
            ai.inpaint_image(small_image, mask, prompt="A beach sunset")

    @patch("parallax_maker.stabilityai.requests.post")
    def test_upscale_image_success(self, mock_post):
        image = Image.new("RGB", (100, 100))
        output_image = Image.new("RGB", (200, 200))
        image_data = BytesIO()
        output_image.save(image_data, format="PNG")
        image_data.seek(0)

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = image_data.read()
        mock_post.return_value = mock_response

        ai = StabilityAI(api_key="fake_api_key")
        result_image = ai.upscale_image(image, prompt="Upscale this image")

        self.assertIsInstance(result_image, Image.Image)
        self.assertEqual(result_image.size, (200, 200))

    @patch("parallax_maker.stabilityai.requests.post")
    def test_upscale_image_invalid_size(self, mock_post):
        large_image = Image.new("RGB", (2000, 2000))
        ai = StabilityAI(api_key="fake_api_key")

        with self.assertRaises(ValueError):
            ai.upscale_image(large_image, prompt="Upscale this image")


if __name__ == "__main__":
    unittest.main()



# (c) 2024 Niels Provos

from dash.dependencies import Input, Output, State, ClientsideFunction

from . import constants as C


def make_clientside_callbacks(app):
    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="store_rect_coords"),
        Output(C.STORE_RECT_DATA, "data"),
        Input(C.IMAGE, "src"),
        Input("evScroll", "n_events"),
    )

    app.clientside_callback(
        ClientsideFunction(
            namespace="clientside", function_name="suppress_contextmenu"
        ),
        Output(C.CTR_INPUT_IMAGE, "id"),
        Input(C.CTR_INPUT_IMAGE, "id"),
    )

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="store_current_tab"),
        Output(C.STORE_CURRENT_TAB, "data"),
        Input(C.STORE_CURRENT_TAB, "data"),
    )

    app.clientside_callback(
        ClientsideFunction(
            namespace="clientside", function_name="record_selected_slice"
        ),
        Output(C.STORE_IGNORE, "data", allow_duplicate=True),
        Input(C.STORE_SELECTED_SLICE, "data"),
        prevent_initial_call=True,
    )

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="visualize_point"),
        Output(C.STORE_IGNORE, "data", allow_duplicate=True),
        Input(C.STORE_CLICKED_POINT, "data"),
        prevent_initial_call=True,
    )

    app.clientside_callback(
        ClientsideFunction(
            namespace="clientside", function_name="preview_canvas_clear"
        ),
        Output(C.STORE_IGNORE, "data", allow_duplicate=True),
        Input(C.STORE_CLEAR_PREVIEW, "data"),
        prevent_initial_call=True,
    )

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="show_bounding_box"),
        Output(C.STORE_IGNORE, "data", allow_duplicate=True),
        Input(C.STORE_BOUNDING_BOX, "data"),
        prevent_initial_call=True,
    )



import unittest
from unittest.mock import patch, MagicMock
import numpy as np
from PIL import Image

from .instance import SegmentationModel


class TestSegmentationModel(unittest.TestCase):

    def setUp(self):
        self.model = SegmentationModel("sam")
        self.mock_image = Image.new("RGB", (120, 90), color="white")
        self.mock_point = (50, 50)
        self.mock_points = [(50, 50), (60, 60), (70, 70)]
        self.mock_mask = np.zeros((90, 120), np.uint8)
        self.mock_mask[40:60, 40:60] = 255

        patcher = patch.object(
            SegmentationModel, "segment_image_mask2former", return_value=self.mock_mask
        )
        self.mock_segment_image = patcher.start()
        self.addCleanup(patcher.stop)

        patcher2 = patch.object(
            SegmentationModel,
            "_get_mask_at_point_function",
            return_value=lambda point: self.mock_mask,
        )
        self.mock_get_mask_at_point_function = patcher2.start()
        self.addCleanup(patcher2.stop)

    def test_transform(self):
        transformations = [
            "identity",
            "rotate_90",
            "rotate_180",
            "rotate_270",
            "flip_h",
            "flip_v",
        ]
        for transformation in transformations:
            transformed_image = self.model._transform_image(
                self.mock_image, transformation
            )
            transformed_point = self.model._transform_point(
                self.mock_point, transformation, self.mock_image.size
            )
            self.assertIsInstance(transformed_image, Image.Image)
            self.assertIsInstance(transformed_point, tuple)
            self.assertEqual(len(transformed_point), 2)

    def test_rotate_point_single(self):
        angles = [0, 90, 180, 270]
        image_size = (120, 90)
        for angle in angles:
            rotated_point = self.model._rotate_point(self.mock_point, angle, image_size)
            self.assertIsInstance(rotated_point, tuple)
            self.assertEqual(len(rotated_point), 2)

    def test_transform_point_multiple(self):
        transformations = [
            "identity",
            "rotate_90",
            "rotate_180",
            "rotate_270",
            "flip_h",
            "flip_v",
        ]
        image_size = (120, 90)
        for transformation in transformations:
            rotated_point = self.model._transform_point(
                self.mock_points, transformation, image_size
            )
            self.assertIsInstance(rotated_point, list)
            for point in rotated_point:
                self.assertIsInstance(point, tuple)
                self.assertEqual(len(point), 2)

    def test_inverse_transform(self):
        transformations = [
            "identity",
            "rotate_90",
            "rotate_180",
            "rotate_270",
            "flip_h",
            "flip_v",
        ]
        for transformation in transformations:
            transformed_image = self.model._transform_image(
                self.mock_image, transformation
            )
            inverse_transformed_mask = self.model._inverse_transform(
                np.array(transformed_image), transformation
            )
            self.assertIsInstance(inverse_transformed_mask, np.ndarray)
            it_mask = Image.fromarray(inverse_transformed_mask)
            self.assertEqual(self.mock_image.size, it_mask.size)

    def test_filter_mask(self):
        masks = []
        for _ in range(5):
            m = np.zeros((90, 120), np.uint8)
            m[40:60, 40:60] = 255
            masks.append(Image.fromarray(m))

        filtered_masks = self.model._filter_mask(masks)
        self.assertIsInstance(filtered_masks, list)
        self.assertGreater(len(filtered_masks), 0)

    @patch.object(SegmentationModel, "_get_mask_at_point_function")
    def test_mask_at_point_blended(self, mock_get_mask_at_point_function):
        def mock_mask_function(point):
            width, height = self.model.image.size
            return np.zeros((height, width), np.uint8)

        mock_get_mask_at_point_function.return_value = mock_mask_function

        self.model.image = self.mock_image
        blended_mask = self.model.mask_at_point_blended(self.mock_point)
        self.assertIsInstance(blended_mask, np.ndarray)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock
from PIL import Image
import numpy as np

from .upscaler import Upscaler


class TestUpscaler(unittest.TestCase):
    def setUp(self):
        # Initialize Upscaler object without creating a model.
        self.upscaler = Upscaler(model_name="simple")

        # Create a dummy input image
        self.input_image = Image.new("RGBA", (800, 800))
        # Fill the alpha with 1s - so that the bounding box is the full image
        self.input_image.putalpha(255)

        # Create a dummy upscaled tile
        self.upscaled_tile_dummy = Image.new("RGB", (1, 1))
        self.upscaled_tile_dummy_real = Image.new(
            "RGB", (self.upscaler.tile_size * 2, self.upscaler.tile_size * 2)
        )

    def test_upscale_image_tiled_overlap(self):
        # Mocking the creation of the model to skip loading model and processor
        with patch.object(self.upscaler, "create_model"):
            # Mock the upscale_tile function to return our dummy tile
            with patch.object(
                self.upscaler, "upscale_tile", return_value=self.upscaled_tile_dummy
            ) as mock_upscale_tile:
                # Mock the integrate_tile to ensure integral processing of tiles
                with patch.object(self.upscaler, "integrate_tile") as mock_integrate:
                    # Call the function to test
                    upscaled_image = self.upscaler.upscale_image_tiled(
                        self.input_image, overlap=64
                    )

                    # Check that integrate_tile was called the correct number of times (expected 3x3 grid based on sizes).
                    self.assertEqual(mock_integrate.call_count, 4)
                    mock_upscale_tile.assert_called()

    def test_upscale_image_tiled_no_overlap(self):
        # Mocking the creation of the model to skip loading model and processor
        with patch.object(self.upscaler, "create_model"):
            # Mock the upscale_tile function to return our dummy tile
            with patch.object(
                self.upscaler, "upscale_tile", return_value=self.upscaled_tile_dummy
            ):
                # Mock the integrate_tile to ensure integral processing of tiles
                with patch.object(self.upscaler, "integrate_tile") as mock_integrate:
                    # Call the function to test
                    upscaled_image = self.upscaler.upscale_image_tiled(
                        self.input_image, overlap=0
                    )

                    # Check that integrate_tile was called the correct number of times (expected 2x2 grid based on sizes).
                    self.assertEqual(mock_integrate.call_count, 4)

    def test_upscale_image_tiled_no_mock(self):
        # Mocking the creation of the model to skip loading model and processor
        with patch.object(self.upscaler, "create_model"):
            # Mock the upscale_tile function to return our dummy tile
            with patch.object(
                self.upscaler,
                "upscale_tile",
                return_value=self.upscaled_tile_dummy_real,
            ):
                # Call the function to test
                upscaled_image = self.upscaler.upscale_image_tiled(
                    self.input_image, overlap=64
                )

                # Check that integrate_tile was called the correct number of times (expected 3x3 grid based on sizes).
                self.assertEqual(upscaled_image.size, (2 * 800, 2 * 800))


class TestIntegrateTile(unittest.TestCase):

    def setUp(self):
        # Create a sample tile and image for testing
        self.tile = np.ones((100, 100, 3), dtype=np.uint8) * 255
        self.image = np.zeros((200, 200, 3), dtype=np.uint8)

    def test_integrate_tile_no_overlap(self):
        # Test integrating a tile with no overlap
        Upscaler.integrate_tile(self.tile, self.image, 50, 50, 150, 150, 0, 0, 0)
        self.assertTrue(np.array_equal(self.image[50:150, 50:150], self.tile))

    def test_integrate_tile_left_overlap(self):
        # Test integrating a tile with left overlap
        Upscaler.integrate_tile(self.tile, self.image, 0, 50, 100, 150, 1, 0, 20)
        self.assertTrue(np.all(self.image[50:150, 1:20] > 0))
        self.assertTrue(np.all(self.image[50:150, 20:100] == 255))

    def test_integrate_tile_top_overlap(self):
        # Test integrating a tile with top overlap
        Upscaler.integrate_tile(self.tile, self.image, 50, 0, 150, 100, 0, 1, 20)
        self.assertTrue(np.all(self.image[1:20, 50:150] > 0))
        self.assertTrue(np.all(self.image[20:100, 50:150] == 255))

    def test_integrate_tile_corner_overlap(self):
        # Test integrating a tile with corner overlap
        Upscaler.integrate_tile(self.tile, self.image, 0, 0, 100, 100, 1, 1, 20)
        self.assertTrue(np.all(self.image[2:20, 2:20] > 0))
        self.assertTrue(np.all(self.image[1:20, 20:100] > 0))
        self.assertTrue(np.all(self.image[20:100, 1:20] > 0))
        self.assertTrue(np.all(self.image[20:100, 20:100] == 255))

    def test_integrate_tile_gradient(self):
        # Test if there is a gradient in the overlapping regions
        overlap = 20
        Upscaler.integrate_tile(self.tile, self.image, 0, 0, 100, 100, 1, 1, overlap)

        # Check left overlap gradient
        left_overlap = self.image[:, :overlap]
        self.assertFalse(np.all(left_overlap == left_overlap[0, 0]))
        self.assertTrue(np.all(np.diff(left_overlap, axis=1) >= 0))

        # Check top overlap gradient
        top_overlap = self.image[:overlap, :]
        self.assertFalse(np.all(top_overlap == top_overlap[0, 0]))
        self.assertTrue(np.all(np.diff(top_overlap, axis=0) >= 0))

        # Check corner overlap gradient
        corner_overlap = self.image[:overlap, :overlap]
        self.assertFalse(np.all(corner_overlap == corner_overlap[0, 0]))
        self.assertTrue(np.all(np.diff(corner_overlap, axis=0) >= 0))
        self.assertTrue(np.all(np.diff(corner_overlap, axis=1) >= 0))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# (c) 2024 Niels Provos
#
# Uses a depth map to segment an image into multiple slices.
# These slices can be used to create animated 2.5D effects.
#

import cv2
import numpy as np
import torch
from torchvision.transforms import Compose
import argparse
from pathlib import Path
from PIL import Image

from .utils import torch_get_device, feather_mask
from .depth import DepthEstimationModel

# for exporting a 3d scene
from .gltf import export_gltf
from .camera import Camera
from .slice import ImageSlice


def generate_depth_map(image, model: DepthEstimationModel, progress_callback=None):
    """
    Generate a depth map from the input image using the specified model.

    Args:
        image (numpy.ndarray): The input image.
        model (DepthEstimationModel): The depth estimation model to use.
            Supported models are "midas" and "dinov2".
        progress_callback (callable, optional): A callback function to report progress.

    Returns:
        numpy.ndarray: The grayscale depth map.
    Raises:
        ValueError: If an unknown model is specified.
    """

    depth_map = model.depth_map(image, progress_callback=progress_callback)
    depth_map = cv2.normalize(depth_map, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)
    return depth_map


def analyze_depth_histogram(depth_map, num_slices=5):
    """Analyze the histogram of the depth map and determine thresholds for segmentation."""

    def calculate_thresholds(depth_map, num_slices):
        thresholds = [0]
        hist, _ = np.histogram(depth_map.flatten(), 256, [0, 256])
        total_pixels = depth_map.shape[0] * depth_map.shape[1]
        target_pixels_per_slice = float(total_pixels) / (num_slices + 1)
        total_sum = 0
        for i in range(1, 256):
            total_sum += hist[i]
            if total_sum >= target_pixels_per_slice * len(thresholds) or i == 255:
                thresholds.append(i)
        return thresholds

    # this is a terrible hack to make sure we get the right number of thresholds
    thresholds = calculate_thresholds(depth_map, num_slices - 1)
    if len(thresholds) != num_slices + 1:
        thresholds = calculate_thresholds(depth_map, num_slices)
    assert (
        len(thresholds) == num_slices + 1
    ), f"Expected {num_slices + 1} thresholds, got {len(thresholds)}"
    return thresholds


def generate_simple_thresholds(depth_map, num_slices=5):
    """Generate simple thresholds based on the depth map."""
    thresholds = [0]
    depth_min = depth_map.min()
    depth_max = depth_map.max()
    depth_range = depth_max - depth_min
    step = depth_range / num_slices
    for i in range(num_slices):
        threshold = int(depth_min + (i + 1) * step)
        thresholds.append(threshold)
    return thresholds


def mask_from_depth(depth_map, threshold_min, threshold_max, prev_mask=None):
    """Generate a mask based on the depth map and thresholds."""
    mask = cv2.inRange(depth_map, threshold_min, threshold_max)
    if prev_mask is not None:
        mask = cv2.bitwise_and(mask, cv2.bitwise_not(prev_mask))
    return mask


def generate_image_slices(image, depth_map, thresholds, num_expand=50):
    """Generate image slices based on the depth map and thresholds, including an alpha channel.

    Args:
        image (numpy.ndarray): The input image.
        depth_map (numpy.ndarray): The depth map corresponding to the input image.
        thresholds (list): A list of threshold values used to segment the depth map.
        num_expand (int, optional): The number of pixels to expand the mask by. Defaults to 50.

    Returns:
        List[ImageSlice]: A list of image slices.

    """
    slices = []

    prev_mask = None
    for i in range(len(thresholds) - 1):
        threshold_min = thresholds[i]
        threshold_max = thresholds[i + 1]

        mask = mask_from_depth(
            depth_map, threshold_min, threshold_max, prev_mask=prev_mask
        )
        masked_image = create_slice_from_mask(image, mask, num_expand)

        image_slice = ImageSlice(image=masked_image, depth=threshold_max)

        slices.append(image_slice)
        prev_mask = mask

    return slices


def create_slice_from_mask(image, mask, num_expand=50):
    """
    Create a slice from an image based on a given mask.

    Args:
        image (PIL.Image.Image or numpy.ndarray): The input image.
        mask (numpy.ndarray): The mask to apply on the image.
        num_expand (int): The number of pixels to expand the mask by.

    Returns:
        numpy.ndarray: The masked image slice.
    """
    if isinstance(image, Image.Image):
        image = np.array(image)

    feathered_mask = feather_mask(mask, num_expand=num_expand)

    # Create a 4-channel image (RGBA)
    masked_image = cv2.cvtColor(image, cv2.COLOR_RGB2RGBA)

    # Set alpha channel values based on the feathered mask
    masked_image[:, :, 3] = feathered_mask
    return masked_image


def render_view(image_slices, camera_matrix, card_corners_3d_list, camera_position):
    """
    Render the current view of the camera.

    Args:
        image_slices (list): A list of image slices.
        camera_matrix (numpy.ndarray): The camera matrix.
        card_corners_3d_list (list): A list of 3D card corners.
        camera_position (numpy.ndarray): The current camera position.

    Returns:
        numpy.ndarray: The rendered image.
    """
    # Start with a blank image with an alpha channel
    rendered_image = np.zeros(
        (image_slices[0].image.shape[0], image_slices[0].image.shape[1], 4),
        dtype=np.uint8,
    )
    rendered_image[:, :, 3] = 1

    for i, slice_image in enumerate(image_slices):
        # Transform the card corners based on the camera position
        rvec = np.zeros((3, 1), dtype=np.float32)  # rotation vector
        tvec = -camera_position.reshape(3, 1)
        card_corners_2d, _ = cv2.projectPoints(
            card_corners_3d_list[i], rvec, tvec, camera_matrix, None
        )
        card_corners_2d = np.int32(card_corners_2d.reshape(-1, 2))

        # Warp the image slice based on the card corners
        cur_image = slice_image.image
        warped_slice = cv2.warpPerspective(
            cur_image,
            cv2.getPerspectiveTransform(
                np.float32(
                    [
                        [0, 0],
                        [cur_image.shape[1], 0],
                        [cur_image.shape[1], cur_image.shape[0]],
                        [0, cur_image.shape[0]],
                    ]
                ),
                np.float32(card_corners_2d),
            ),
            (rendered_image.shape[1], rendered_image.shape[0]),
        )

        # Alpha Compositing of the warped slice with the rendered image
        blend_with_alpha(rendered_image, warped_slice)

    return rendered_image


# XXX - consider whether this should return the image with the alpha patch instead of just the alpha
def remove_mask_from_alpha(image, mask):
    """
    Removes the masked region from the alpha channel of an image.

    Args:
        image (numpy.ndarray): The input image with an alpha channel.
        mask (numpy.ndarray): The mask indicating the region to be removed.

    Returns:
        numpy.ndarray: The modified image with the masked region removed from the alpha channel.
    """
    assert image.shape[2] == 4, "Image must have an alpha channel"
    assert (
        image.shape[:2] == mask.shape
    ), f"Image and mask must have the same dimensions: {image.shape[:2]} vs {mask.shape}"

    inverted_mask = 1 - mask / 255.0
    slice_mask = image[:, :, 3] / 255.0

    final_mask = inverted_mask * slice_mask
    final_mask = (final_mask * 255).astype(np.uint8)
    final_mask = np.clip(final_mask, 0, 255)

    return final_mask


def blend_with_alpha(target_image, merge_image):
    """
    Blends the merge_image with the target_image using alpha blending.

    Parameters:
    target_image (numpy.ndarray): The target image to blend with.
    merge_image (numpy.ndarray): The image to be merged with the target image.

    Returns:
    None
    """
    alpha = merge_image[:, :, 3] / 255.0
    target_image[:, :, 0] = (1 - alpha) * target_image[:, :, 0] + alpha * merge_image[
        :, :, 0
    ]
    target_image[:, :, 1] = (1 - alpha) * target_image[:, :, 1] + alpha * merge_image[
        :, :, 1
    ]
    target_image[:, :, 2] = (1 - alpha) * target_image[:, :, 2] + alpha * merge_image[
        :, :, 2
    ]
    target_image[:, :, 3] = np.maximum(target_image[:, :, 3], merge_image[:, :, 3])


def render_image_sequence(
    output_path,
    image_slices,
    card_corners_3d_list,
    camera_matrix,
    camera_position,
    push_distance=100,
    num_frames=100,
    progress_callback=None,
):
    """
    Renders a sequence of images with varying camera positions.

    Args:
        output_path (str): The path to the output directory where the rendered images will be saved.
        image_slices (list): A list of image slices.
        card_corners_3d_list (list): A list of 3D card corners.
        camera_matrix (numpy.ndarray): The camera matrix.
        camera_position (list): The initial camera position.

    Returns:
        None
    """

    if progress_callback:
        progress_callback(0, num_frames)

    output_path = Path(output_path)
    for i in range(num_frames):
        # Update the camera position
        camera_position[2] += float(push_distance) / num_frames

        # Render the view
        rendered_image = render_view(
            image_slices, camera_matrix, card_corners_3d_list, camera_position
        )

        image_name = f"rendered_image_{i:03d}.png"
        output_image_path = output_path / image_name

        cv2.imwrite(
            str(output_image_path), cv2.cvtColor(rendered_image, cv2.COLOR_RGBA2BGR)
        )

        if progress_callback:
            progress_callback(i + 1, num_frames)


def process_image(
    image_path,
    output_path,
    num_slices=5,
    use_simple_thresholds=False,
    create_depth_map=True,
    create_image_slices=True,
    create_image_animation=True,
    push_distance=100,
    depth_model="midas",
):
    """
    Process the input image to generate a depth map and image slices.

    Args:
        image_path (str): The path to the input image file.
        output_path (str): The path to the output directory where the generated files will be saved.
        num_slices (int, optional): The number of image slices to generate. Defaults to 5.
        use_simple_thresholds (bool, optional): Whether to use simple thresholds for image slices. Defaults to False.
        create_depth_map (bool, optional): Whether to generate the depth map. Defaults to True.
        create_image_slices (bool, optional): Whether to generate the image slices. Defaults to True.

    Returns:
        None
    """
    # Load the input image
    image = cv2.imread(image_path)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    print("Image shape:", image.shape)

    output_path = Path(output_path)
    depth_map_path = output_path / "depth_map.png"
    if create_depth_map:
        # Generate the depth map
        model = DepthEstimationModel(depth_model)
        depth_map = generate_depth_map(image, model)

        # save the depth map to a file
        cv2.imwrite(str(depth_map_path), depth_map)
    else:
        # Load the depth map
        depth_map = cv2.imread(str(depth_map_path), cv2.IMREAD_GRAYSCALE)

    if use_simple_thresholds:
        thresholds = generate_simple_thresholds(depth_map, num_slices=num_slices)
    else:
        thresholds = analyze_depth_histogram(depth_map, num_slices=num_slices)

    if create_image_slices:
        # Generate image slices
        image_slices = generate_image_slices(image, depth_map, thresholds)

        # Save the image slices
        for i, slice_image in enumerate(image_slices):
            output_image_path = output_path / f"image_slice_{i}.png"
            slice_image.filename = output_image_path
            print(f"Saving image slice: {output_image_path}")
            slice_image.save_image()
    else:
        # Load the image slices
        image_slices = []
        for i in range(num_slices):
            input_image_path = output_path / f"image_slice_{i}.png"
            image_slice = ImageSlice(filename=input_image_path)
            print(f"Loading image slice: {input_image_path}")
            image_slice.read_image()
            image_slices.append(slice_image)

    # Set up the camera and cards
    for i, image_slice in enumerate(image_slices):
        image_slice.depth = thresholds[i + 1]

    image_height, image_width, _ = image_slices[0].image.shape
    camera = Camera(100.0, 500.0, 100.0)
    camera_matrix = camera.camera_matrix(image_width, image_height)
    card_corners_3d_list = []
    for i, image_slice in enumerate(image_slices):
        card = image_slice.create_card(image_height, image_width, camera)
        card_corners_3d_list.append(card)

    # Render the initial view
    camera_position = np.array([0, 0, -100], dtype=np.float32)
    rendered_image = render_view(
        image_slices, camera_matrix, card_corners_3d_list, camera_position
    )
    # Display the rendered image
    cv2.imshow("Rendered Image", cv2.cvtColor(rendered_image, cv2.COLOR_RGBA2BGR))
    cv2.waitKey(0)
    cv2.destroyAllWindows()

    if create_image_animation:
        render_image_sequence(
            output_path,
            image_slices,
            card_corners_3d_list,
            camera_matrix,
            camera_position,
            push_distance=push_distance,
        )

    image_paths = [output_path / f"image_slice_{i}.png" for i in range(num_slices)]

    output_path = Path(output_path) / "model.gltf"
    export_gltf(output_path, camera, image_slices, image_paths)


def main():
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description="Process image and generate image slices."
    )
    parser.add_argument("-i", "--image", type=str, help="Path to the input image")
    parser.add_argument(
        "-o", "--output", type=str, default=".", help="Path to the output directory"
    )
    parser.add_argument(
        "-n",
        "--num_slices",
        type=int,
        default=5,
        help="Number of image slices to generate",
    )
    parser.add_argument(
        "-s",
        "--use_simple_thresholds",
        action="store_true",
        help="Use simple thresholds for image slices",
    )
    parser.add_argument(
        "-d",
        "--skip_depth_map",
        action="store_true",
        help="Skip generating the depth map",
    )
    parser.add_argument(
        "-g",
        "--skip_image_slices",
        action="store_true",
        help="Slip generating the image slices",
    )
    parser.add_argument(
        "-a",
        "--skip_image_animation",
        action="store_true",
        help="Skip generating the animated images",
    )
    parser.add_argument(
        "-p",
        "--push_distance",
        type=int,
        default=100,
        help="Distance to push the camera in the animation",
    )
    parser.add_argument(
        "--depth_model",
        type=str,
        default="midas",
        help="Depth model to use (midas or zoedepth). Default is midas and tends to work better.",
    )
    args = parser.parse_args()

    # Check if image path is provided
    if args.image:
        # Call the function with the image path
        use_simple_thresholds = args.use_simple_thresholds
        generate_depth_map = not args.skip_depth_map
        generate_image_slices = not args.skip_image_slices
        generate_image_animation = not args.skip_image_animation
        process_image(
            args.image,
            args.output,
            num_slices=args.num_slices,
            use_simple_thresholds=use_simple_thresholds,
            create_depth_map=generate_depth_map,
            create_image_slices=generate_image_slices,
            create_image_animation=generate_image_animation,
            push_distance=args.push_distance,
            depth_model=args.depth_model,
        )
    else:
        print("Please provide the path to the input image using --image or -i option.")


if __name__ == "__main__":
    main()
